/***************************************************************************
 *   Copyright (C) 2007 by BogDan Vatra                                    *
 *   bogdan@licentia.eu                                                    *
 *                                                                         *
 *   This program is free software: you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/
#include <QVariant>
#include <QBuffer>
#include <QDir>
#include <QCoreApplication>
#include <QPluginLoader>
#include <QDebug>
#include <QDate>
#include <QDateTime>
#include <QTime>
#include <QKeySequence>
#include <QMetaProperty>
#include <QBuffer>

#include "paintdevice.h"
#include "paintengine.h"
#include "globals.h"

namespace Report
{

PaintEngine::PaintEngine(PaintEngineFeatures caps)
		: QPaintEngine(caps), m_document(0)
{
}


bool PaintEngine::begin(QPaintDevice *pdev)
{
	m_paintDevice = dynamic_cast<PaintDevice *>(pdev);
	return m_paintDevice;
}

void PaintEngine::setPageNode(QDomNode & pageNode)
{
	m_pageNode = pageNode;
}

void PaintEngine::setDomDocument(QDomDocument* document)
{
	m_document = document;
}

bool PaintEngine::end()
{
	if (!m_paintDevice)
		return false;
	m_paintDevice->endDoc();
	return true;
}

void PaintEngine::updateState(const QPaintEngineState &state)
{
	if (state.state() & QPaintEngine::DirtyBackground)
		m_backgroundBrush = state.backgroundBrush();

	if (state.state() & QPaintEngine::DirtyBackgroundMode)
		m_backgroundMode = state.backgroundMode();

	if (state.state() & QPaintEngine::DirtyBrush)
		m_brush = state.brush();

	if (state.state() & QPaintEngine::DirtyBrushOrigin)
		m_brushOrigin = state.brushOrigin();

	if (state.state() & (QPaintEngine::DirtyClipRegion | QPaintEngine::DirtyClipPath | QPaintEngine::DirtyClipEnabled))
		m_clipOperation = state.clipOperation();

	if (state.state() & (QPaintEngine::DirtyClipRegion | QPaintEngine::DirtyClipPath | QPaintEngine::DirtyClipEnabled))
		m_clipRect = state.clipRegion().boundingRect();

	if (state.state() & QPaintEngine::DirtyCompositionMode)
		m_compositionMode = state.compositionMode();

	if (state.state() & QPaintEngine::DirtyFont)
		m_font = state.font();

	if (state.state() & QPaintEngine::DirtyClipEnabled)
		m_isClipEnabled = state.isClipEnabled();

	if (state.state() & QPaintEngine::DirtyTransform)
		m_transform = state.transform();

	if (state.state() & QPaintEngine::DirtyOpacity)
		m_opacity = state.opacity();

	if (state.state() & QPaintEngine::DirtyPen)
		m_pen = state.pen();

	if (state.state() & QPaintEngine::DirtyHints)
		m_renderHints = state.renderHints();

	m_state |= state.state();

}

void PaintEngine::drawState()
{
	if (!m_state)
		return;

	QDomElement el = m_document->createElement("stateChanged");

	if (m_state & QPaintEngine::DirtyTransform)
		el.appendChild(variatToDom(m_document, "transform", m_transform));

	if (m_state & QPaintEngine::DirtyHints)
		el.appendChild(variatToDom(m_document, "renderHints", (int)m_renderHints));

	if (m_state & QPaintEngine::DirtyBackground)
		el.appendChild(variatToDom(m_document, "backgroundBrush", m_backgroundBrush));

	if (m_state & QPaintEngine::DirtyBackgroundMode)
		el.appendChild(variatToDom(m_document, "backgroundMode", m_backgroundMode));

	if (m_state & QPaintEngine::DirtyBrush)
		el.appendChild(variatToDom(m_document, "brush", m_brush));

	if (m_state & QPaintEngine::DirtyBrushOrigin)
		el.appendChild(variatToDom(m_document, "brushOrigin", m_brushOrigin));

	if (m_state & QPaintEngine::DirtyCompositionMode)
		el.appendChild(variatToDom(m_document, "compositionMode", m_compositionMode));

	if (m_state & QPaintEngine::DirtyFont)
		el.appendChild(variatToDom(m_document, "font", m_font));

	if (m_state & QPaintEngine::DirtyOpacity)
		el.appendChild(variatToDom(m_document, "opacity", m_opacity));

	if (m_state & QPaintEngine::DirtyPen)
		el.appendChild(variatToDom(m_document, "pen", m_pen));

	if (m_state & (QPaintEngine::DirtyClipRegion | QPaintEngine::DirtyClipPath | QPaintEngine::DirtyClipEnabled))
		el.appendChild(variatToDom(m_document, "clipOperation", m_clipOperation));

	m_clipRect.setWidth(m_pen.width() + m_clipRect.width());
	m_clipRect.setHeight(m_pen.width() + m_clipRect.height());

	if (m_state & (QPaintEngine::DirtyClipRegion | QPaintEngine::DirtyClipPath | QPaintEngine::DirtyClipEnabled))
		el.appendChild(variatToDom(m_document, "clipRegion", m_clipRect));

	if (m_state & QPaintEngine::DirtyClipEnabled)
		el.appendChild(variatToDom(m_document, "isClipEnabled", m_isClipEnabled));
	m_pageNode.appendChild(el);

	m_state = 0;
}

void PaintEngine::drawRects(const QRect *rects, int rectCount)
{
	drawState();
	for (int i = 0;i < rectCount;i++)
		m_pageNode.appendChild(variatToDom(m_document, "rect", QRectF(rects[i].x() - m_transform.dx(), rects[i].y() - m_transform.dy(), rects[i].width(), rects[i].height())));
}

void PaintEngine::drawRects(const QRectF *rects, int rectCount)
{
	drawState();
	for (int i = 0;i < rectCount;i++)
		m_pageNode.appendChild(variatToDom(m_document, "rect", QRectF(rects[i].x() - m_transform.dx(), rects[i].y() - m_transform.dy(), rects[i].width(), rects[i].height())));
}

void PaintEngine::drawLines(const QLine *lines, int lineCount)
{
	drawState();
	for (int i = 0;i < lineCount;i++)
		m_pageNode.appendChild(variatToDom(m_document, "line", QLineF(lines[i].x1() - m_transform.dx(), lines[i].y1() - m_transform.dy(), lines[i].x2() - m_transform.dx(), lines[i].y2() - m_transform.dy())));
}

void PaintEngine::drawLines(const QLineF *lines, int lineCount)
{
	drawState();
	for (int i = 0;i < lineCount;i++)
		m_pageNode.appendChild(variatToDom(m_document, "line", QLineF(lines[i].x1() - m_transform.dx(), lines[i].y1() - m_transform.dy(), lines[i].x2() - m_transform.dx(), lines[i].y2() - m_transform.dy())));
}

void PaintEngine::drawEllipse(const QRectF &r)
{
	drawState();
	m_pageNode.appendChild(variatToDom(m_document, "ellipse", QRectF(r.x() - m_transform.dx(), r.y() - m_transform.dy(), r.width(), r.height())));
}
void PaintEngine::drawEllipse(const QRect &r)
{
	drawState();
	m_pageNode.appendChild(variatToDom(m_document, "ellipse", QRectF(r.x() - m_transform.dx(), r.y() - m_transform.dy(), r.width(), r.height())));
}

void PaintEngine::drawPath(const QPainterPath &path)
{
	Q_UNUSED(path);
#ifndef WIN32
#warning TODO: drawPath
#endif
	drawState();
	qDebug() << "drawPath is not finished";
}

void PaintEngine::drawPoints(const QPointF *points, int pointCount)
{
	drawState();
	for (int i = 0;i < pointCount;i++)
		m_pageNode.appendChild(variatToDom(m_document, "point", points[i]));
}

void PaintEngine::drawPoints(const QPoint *points, int pointCount)
{
	drawState();
	for (int i = 0;i < pointCount;i++)
		m_pageNode.appendChild(variatToDom(m_document, "point", points[i]));
}

void PaintEngine::drawPolygon(const QPointF *points, int pointCount, PolygonDrawMode mode)
{
	drawState();
	if (QPaintEngine::PolylineMode != mode)
	{
		qDebug() << "drawPolygon is not fully implemented. Only PolylineMode is working";
		return;
	}

	for (int i = 1;i < pointCount;i++)
	{
		QLineF line = QLineF(points[i-1].x(), points[i-1].y(), points[i].x(), points[i].y());
		drawLines(&line, 1);
	}
}

void PaintEngine::drawPolygon(const QPoint *points, int pointCount, PolygonDrawMode mode)
{
	drawState();
	if (QPaintEngine::PolylineMode != mode)
	{
		qDebug() << "drawPolygon is not fully implemented. Only PolylineMode is working";
		return;
	}

	for (int i = 1;i < pointCount;i++)
	{
		QLine line = QLine(points[i-1].x(), points[i-1].y(), points[i].x(), points[i].y());
		drawLines(&line, 1);
	}
}

void PaintEngine::drawPixmap(const QRectF &r, const QPixmap &pm, const QRectF &sr)
{
	drawState();
	QRectF rr=r;
	rr.translate(-m_transform.dx(),-m_transform.dy());
	QDomElement el = m_document->createElement("pixmap");
	el.appendChild(variatToDom(m_document, "toRect", rr));
	el.appendChild(variatToDom(m_document, "sourceRect", sr));
	el.appendChild(variatToDom(m_document, "img", pm));
	m_pageNode.appendChild(el);
}

void PaintEngine::drawTextItem(const QPointF &p, const QTextItem &textItem)
{
	drawState();
	QDomElement el = m_document->createElement("text");
	el.appendChild(variatToDom(m_document, "toPoint", p));
	el.appendChild(variatToDom(m_document, "font", textItem.font()));
	el.appendChild(variatToDom(m_document, "renderFlags", (int)textItem.renderFlags()));
	el.appendChild(variatToDom(m_document, "width", textItem.width()));
	el.appendChild(variatToDom(m_document, "ascent", textItem.ascent()));
	el.appendChild(variatToDom(m_document, "descent", textItem.descent()));
	el.appendChild(variatToDom(m_document, "string", textItem.text()));
	m_pageNode.appendChild(el);
}

void PaintEngine::drawTiledPixmap(const QRectF &r, const QPixmap &pm, const QPointF &s)
{
	drawState();
	QRectF rr=r;
	rr.translate(-m_transform.dx(),-m_transform.dy());
	QDomElement el = m_document->createElement("tiledPixmap");
	el.appendChild(variatToDom(m_document, "toRect", rr));
	el.appendChild(variatToDom(m_document, "toPoint", s));
	el.appendChild(variatToDom(m_document, "img", pm));
	m_pageNode.appendChild(el);
}

void PaintEngine::drawImage(const QRectF &r, const QImage &pm, const QRectF &sr,
                            Qt::ImageConversionFlags flags)
{
	drawState();
	QRectF rr=r;
	rr.translate(-m_transform.dx(),-m_transform.dy());
	QDomElement el = m_document->createElement("image");
	el.appendChild(variatToDom(m_document, "toRect", rr));
	el.appendChild(variatToDom(m_document, "sourceRect", sr));
	el.appendChild(variatToDom(m_document, "flags", (int)flags));
	el.appendChild(variatToDom(m_document, "img", pm));
	m_pageNode.appendChild(el);
}

QPaintEngine::Type PaintEngine::type() const
{
	return QPaintEngine::User;
}

PaintEngine::~PaintEngine()
{
}

}
